#!/usr/bin/env python3
# Copyright 2016, 2018 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import
from __future__ import print_function

import errno
import glob
import gzip
import os
import shutil
import time
from collections import defaultdict, namedtuple

import redis
from datetime import datetime
from redis import ConnectionError

__BACKUP_DIR = '/var/backup'
__DATA_DIR = '/mnt/data/redis/master'
__RDB = 'dump.rdb'
__AOF = 'appendonly.aof'
__MB = 2 ** 20
__BUF_SIZE = 128 * __MB

__DEFAULT_SAVEPOINTS = '900 1 300 10 60 10000'
__ABSOLUTE_MIN_DISK = 300 * 1024 * 1024  # Won't go less than this  disk
__MAX_FILES = 7

__HOST = '127.0.0.1'
__PORT = 6379


def log(*args):
    ts = datetime.now().strftime('%Y-%m-%d_%H%M%S')
    print('cts-backup', ts, *args)


def get_info():
    try:
        return redis.StrictRedis(__HOST, __PORT).info()
    except ConnectionError as ce:
        log("REDIS connection error %s" % (ce))
    except redis.exceptions.RedisError as re:
        log("REDIS error %s" % (re))
    except OSError as ose:
        log("OS error %s" % (ose))


def get_latest_activity():
    info = get_info()
    return info.get('persistence') if info else None


def check_minimum_disk(backup_dir, min_allowed_space=__ABSOLUTE_MIN_DISK):
    min_allowed_space = min_allowed_space if min_allowed_space > __ABSOLUTE_MIN_DISK else __ABSOLUTE_MIN_DISK
    _, _, available = shutil.disk_usage(backup_dir)

    # Removed forced AOF rewrite on theory it would topple constrained disk

    try:
        conn = redis.StrictRedis(__HOST, __PORT)
        # Pre-emptively shut off point-in-time
        if available < min_allowed_space:
            log('Stopping point-in-time rdb backup - available(%s) required(%s)' % (available, min_allowed_space))
            conn.config_set('SAVE', '')
        else:
            # Check to see if someone else set savepoints
            current = conn.config_get('SAVE')
            if not current['save']:
                conn.config_set('SAVE', __DEFAULT_SAVEPOINTS)
    except ConnectionError as ce:
        log("REDIS connection error %s" % (ce))
    except redis.exceptions.RedisError as re:
        log("REDIS error %s" % (re))


def cleanup_data(backup_dir, pattern):
    '''
    :param backup_dir:
    :param pattern:pattern for glob to use find backups 'aof' or 'rdb'
    :return:
    '''

    files = [(f, os.path.getmtime(f), os.path.getsize(f)) for f in glob.iglob(os.path.join(backup_dir, pattern))]
    if not files:
        log("No available files in %s matching %s to purge" % (backup_dir, pattern))
        return

    for fs in files:
        remove_backup(fs[0])
        log("Removed %s" % fs[0])


def remove_backup(src):
    try:
        os.remove(src)
    except OSError as ose:
        log('Could not remove backup %s %s' % (src, ose))


def cleanup_backups(backup_dir, pattern, min_allowed_space=__ABSOLUTE_MIN_DISK, max_files=__MAX_FILES):
    '''
    :param backup_dir:
    :param pattern:pattern for glob to use find backups 'aof' or 'rdb'
    :param min_allowed_space:minimum_disk_space - 300M
    :param sliding_window:how old files need to be to be purged - 24 hrs
    :param max_files:max # of files 7
    :return:
    '''

    File = namedtuple('File', 'name timestamp size')
    files = [File(f, os.path.getmtime(f), os.path.getsize(f)) for f in glob.iglob(os.path.join(backup_dir, pattern))]
    if not files:
        log("No available files in %s matching %s to purge" % (backup_dir, pattern))
        return

    files.sort(key=lambda x: x.timestamp)
    if len(files) > max_files:
        for f in files[:len(files) - max_files]:
            remove_backup(f.name)
            log('Removed:bytime:{}:{}'.format(f.timestamp, f.name))
        files = files[len(files) - max_files:]

    # Still need some space?
    min_allowed_space = min_allowed_space if min_allowed_space > __ABSOLUTE_MIN_DISK else __ABSOLUTE_MIN_DISK

    # If disk space getting low, purge oldest until above minimum threshold
    _, _, available = shutil.disk_usage(backup_dir)
    while files and available < min_allowed_space:
        f = files.pop(0)
        remove_backup(f.name)
        log('Removed:overspace:{}:{}:{}({})'.format(min_allowed_space, (available - f.size), f.name, f.size))
        _, _, available = shutil.disk_usage(backup_dir)


def get_mtime(file):
    try:
        return os.path.getmtime(os.path.realpath(file))
    except OSError as ose:
        log("could not get mtime for file %s because %s" % (file, ose))
    return 0


def backup(kind, data_path, backup_path, symlink_path, force=False):
    if not os.path.exists(data_path):
        log("No valid datafile to backup:%s" % (data_path))
        return

    last_aof_status = role = 'unknown'
    info = get_info()
    if info:
        role = 'master' if info['role'] == 'master' else 'replica'
        last_aof_status = info['aof_last_write_status']

        if info['loading']:
            log("Redis currently reloading, skipping backup %s" % (data_path))
            return

    # TODO: dedup
    if not force:
        if kind == 'aof' and last_aof_status != 'ok':
            log("Skipping backup. AOF backup status %s for %s" % (last_aof_status, data_path))
            return

    lnm = os.path.basename(data_path) + '_' + role + '_' + 'latest.gz'
    link_mtime = get_mtime(os.path.join(symlink_path, lnm))
    data_mtime = get_mtime(data_path)

    if data_mtime > link_mtime:
        backup_file(role, data_path, backup_path, symlink_path)


def symlink(src, dst):
    try:
        os.symlink(src, dst)
    except OSError as ose:
        if ose.errno == errno.EEXIST:
            remove_backup(dst)
            try:
                os.symlink(src, dst)
            except OSError as ex:
                log('Could not symlink %s %s %s' % (src, dst, ex))
        else:
            log('Could not symlink %s %s %s' % (src, dst, ose))


def backup_file(role, data_path, backup_path, symlink_path, min_allowed_space=__ABSOLUTE_MIN_DISK):
    _, _, available = shutil.disk_usage(backup_path)
    file_sz = os.path.getsize(data_path)

    if available - file_sz < min_allowed_space:
        log("Limited diskspace(%s) - require(%s) cannot backup size(%s) %s" % (
        available, min_allowed_space, file_sz, data_path))
        return

    fmt = os.path.basename(data_path) + '_' + role + '_' + '%Y-%m-%d_%H%M%S.gz'
    lnm = os.path.basename(data_path) + '_' + role + '_' + 'latest.gz'

    fname = datetime.fromtimestamp(os.path.getmtime(data_path)).strftime(fmt)
    backup = os.path.join(backup_path, fname)

    with open(data_path, 'rb', buffering=__BUF_SIZE) as inp:
        with gzip.open(backup, 'wb') as outp:
            shutil.copyfileobj(inp, outp, length=__BUF_SIZE)
            outp.flush()
        inp.flush()

    symlink(backup, os.path.join(symlink_path, lnm))
    log("Backed up %s to %s and %s" % (data_path, backup, os.path.join(symlink_path, lnm)))


def backup_aof():
    cleanup_backups(__BACKUP_DIR, 'dump.rdb*.gz')
    cleanup_backups(__BACKUP_DIR, 'appendonly.aof*.gz')
    backup('aof', '/mnt/data/redis/master/appendonly.aof', '/var/backup', '/var/log/backup', force=True)


def clean_start():
    cleanup_data(__DATA_DIR, '*dump.rdb*')
    cleanup_data(__DATA_DIR, '*appendonly.aof*')
    cleanup_data(__DATA_DIR, '*rewriteaof*')


def main():
    cleanup_backups(__BACKUP_DIR, 'dump.rdb*.gz')
    cleanup_backups(__BACKUP_DIR, 'appendonly.aof*.gz')
    backup('aof', '/mnt/data/redis/master/appendonly.aof', '/var/backup',
           '/var/log/backup')
    backup('rdb', '/mnt/data/redis/master/dump.rdb', '/var/backup',
           '/var/log/backup')
    # check_minimum_disk('/var/backup')


if __name__ == "__main__":
    main()
