#!/usr/bin/env python3
# Copyright 2016, 2018 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import
from __future__ import print_function

import subprocess
from logging import Formatter, getLogger, INFO
from logging.handlers import SysLogHandler

import argparse
import redis
from datetime import datetime

from cts_backup.redis_backup import backup_aof, clean_start

__NETQ_APP_SVC = 'netq-appliance'
__HOST = '127.0.0.1'
__PORT = 6379


def get_logger(proc_name):
    logger = getLogger(proc_name)
    if len(logger.handlers):
        # Logger already configured
        return logger

    logging_level = INFO

    handler = SysLogHandler(address='/dev/log', facility=SysLogHandler.LOG_USER)
    formatter = Formatter(proc_name + '[%(process)d]: %(levelname)s: %(message)s')
    handler.setLevel(logging_level)
    handler.setFormatter(formatter)
    logger.setLevel(logging_level)
    logger.addHandler(handler)
    return logger


def log(*args):
    ts = datetime.now().strftime('%Y-%m-%d_%H%M%S')
    print('cts-backup', ts, *args)
    get_logger('cts-redis-reset').info(*args)


def manage_svc(svcname, cmd):
    systemctl = ['sudo', '/bin/systemctl', cmd, svcname]

    if cmd in ['start', 'stop', 'status']:
        try:
            subprocess.check_output(systemctl)
            print('%s %s ... Success!' % (cmd, svcname))
        except subprocess.CalledProcessError:
            print('Failed to %s %s' % (cmd, svcname))


def redis_wrap(fcn, *args, **kwargs):
    try:
        conn = redis.StrictRedis(host=__HOST, port=__PORT, socket_timeout=.5)
        return fcn(conn, *args, **kwargs)
    except (redis.ConnectionError, redis.exceptions.RedisError, OSError) as ce:
        raise Exception(str(ce))


def get_config(cfg):
    return redis_wrap(lambda conn: conn.config_get(cfg))


def set_config(cfg, val):
    return redis_wrap(lambda conn: conn.config_set(cfg, val))


def save_config():
    return redis_wrap(lambda conn: conn.config_rewrite())


def get_info():
    return redis_wrap(lambda conn: conn.info())


def get_maxmemory():
    return int(get_config('maxmemory')['maxmemory'])


def appendonly(start=True):
    cmd = 'yes' if start else 'no'
    return set_config('APPENDONLY', cmd)


def save_off():
    return set_config('SAVE', '')


def snapshot():
    return redis_wrap(lambda conn: conn.execute_command('BGREWRITEAOF'))


def redis_down():
    # TODO: Only shutdown REDIS
    manage_svc(__NETQ_APP_SVC, 'stop')


def redis_up():
    manage_svc(__NETQ_APP_SVC, 'start')


def clean_sweep():
    redis_down()
    clean_start()
    redis_up()


def reset():
    snapshot()
    appendonly(start=False)
    save_off()
    clean_sweep()
    backup_aof()
    appendonly(start=True)
    save_config()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--clean', action='store_true', help='Restart REDIS clean without backups.')
    parser.add_argument('--reset', action='store_true', help='Snapshot REDIS. Backup. Clean.')
    args = parser.parse_args()

    if args.reset:
        reset()
    elif args.clean:
        clean_sweep()


if __name__ == '__main__':
    main()
